#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#ifdef _WIN32

// WINDOZE
#include "winsock2.h"

#include "WS2TCPIP.h"
#else

// Linux
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <netdb.h>

#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>


#endif


#define STRING_SIZE				100
#define MAX_SEQUENCE			4
#define KNOCK_TIME_WINDOW		3000
#define MIN_SEQUENCE_KNOCKS		0
#define MAX_SEQUENCE_KNOCKS		9

#define DEFAULTFILENAME			"knocksettings.txt"

/*****************************************************************************\
*
\*****************************************************************************/


/*****************************************************************************\
*
\*****************************************************************************/

#ifdef _WIN32

struct ip 
{ 
	u_char  ip_hl:4;        /* header length */ 
	u_char  ip_v:4;         /* version */ 
	u_char  ip_tos;         /* type of service */ 
	u_short ip_len;         /* total length */ 
	u_short ip_id;			/* identification */ 
	u_short ip_off;         /* fragment offset field */ 
	u_char  ip_ttl;         /* time to live */ 
	u_char  ip_p;           /* protocol */ 
	u_short ip_sum;			/* checksum */ 
	struct  in_addr ip_src; /* source address */ 
	struct  in_addr ip_dst; /* destination address */ 
};

struct tcphdr
{
    u_short source;
    u_short dest;
    u_int   seq;
    u_int   ack_seq;

    u_short res1:4;
    u_short doff:4;
    u_short fin:1;
    u_short syn:1;
    u_short rst:1;
    u_short psh:1;
    u_short ack:1;
    u_short urg:1;
    u_short res2:2;

    u_short window;
    u_short check;
    u_short urg_ptr;
};


#define ICMP_ECHOREPLY	0	/* ICMP type: echo reply */
#define ICMP_ECHO		8   /* ICMP type: echo request */
/* definition of ICMP header as per RFC 792 */ 
struct icmp 
{ 
  u_char  icmp_type;      /* type of message */ 
  u_char  icmp_code;      /* type sub code */ 
  u_short icmp_cksum;    /* ones complement cksum */ 
  u_short icmp_id;       /* identifier */ 
  u_short icmp_seq;      /* sequence number */ 
  char icmp_data[1];     /* data */ 
}; 

//#define ICMP_HDR_LEN sizeof(ICMP_HDR)

typedef int socklen_t;


#else

typedef int SOCKET;

#endif


/*****************************************************************************\
*
\*****************************************************************************/

unsigned long 	dotIpAddress;
SOCKET 			sock;
struct hostent	*lpHost;
sockaddr_in		service;
int				result;
int             rawSocket;
int				dgramSocket;

#ifdef _WIN32
WSADATA			wsaData;
#endif

char		destinationPc[STRING_SIZE];
int			destinationPort;
int			knocks[MAX_SEQUENCE];
int			knockTimeWindow;
  
/*****************************************************************************\
*
*
\*****************************************************************************/
unsigned short csum (unsigned short *buf, int nwords)
{
	unsigned long sum;
	int i;

	i=0;
	sum=0;

	while (i<nwords)
	{
		sum+= buf[i];
		i++;
	}

	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	sum = ~sum;
	return  (unsigned short)sum;
}



/*****************************************************************************\
*
*
\*****************************************************************************/
  
void printError()
{
#ifdef _WIN32
	int error;
	
	error=WSAGetLastError();
	printf("Error: %d ", error);
	switch (error)
	{
	case WSAEINTR:
		printf("Interrupted function call\n");
		break;
	case WSAEACCES:
		printf("Permission denied\n");
		break;
	case WSAEFAULT:
		printf("Invalid address\n");
		break;
	case WSAEINVAL:
		printf("Invalid Argument\n");
		break;
	default:
		printf("Unknown\n");
	}


#else
	perror(NULL);
#endif
}
/*****************************************************************************\
*
*
\*****************************************************************************/
void initialise()
{
    // initialise the library
#ifdef _WIN32

    result=WSAStartup(MAKEWORD(2, 2), &wsaData);
    if (result!=NO_ERROR)
    {
        printf("Error initalising winsock2\n");
    }
#endif
}

/*****************************************************************************\
*
*
\*****************************************************************************/
void wrapUp()
{
#ifdef _WIN32
    WSACleanup();
#endif
}



/*****************************************************************************\
*
*
\*****************************************************************************/
int openDgramSocket(char* address, int port)
{
	struct sockaddr_in 	sin;			// the sockaddr_in containing the dest. address is used
	int					result;

	
	dgramSocket=socket (AF_INET, SOCK_DGRAM, 0);	// open socket 

	printf("Socket opened: %d \n", dgramSocket);
	if (dgramSocket<0)
	{
		printError();
	}


   // finally, it is very advisable to do a IP_HDRINCL call, to make sure
   // that the kernel knows the header is included in the data, and doesn't
   // insert its own header into the packet before our data 
   // This enables source IP Spoofing...


										// in sendto() to determine the datagrams path 
	sin.sin_family      = AF_INET;
	sin.sin_port        = htons (port);						// you byte-order >1byte header values to network
			  												// byte order (not needed on big endian machines)
	sin.sin_addr.s_addr = inet_addr (address);


	result = connect (dgramSocket, (struct sockaddr*)&sin, sizeof(sin)) ;


	if (result<0)
	{
		printError();
	}

	return dgramSocket;
}


/*****************************************************************************\
*
*
\*****************************************************************************/

void closeDgramSocket()
{
#ifdef _WIN32
	// TODO
#else
  close(dgramSocket);
#endif
}


/*****************************************************************************\
*
*
\*****************************************************************************/
void sendDgram(int number)
{
  int					result;
  int					i;
  char					dummy;	 
 
  dummy=5;

  i=0;
  while (i<number)
  {
	result=send(dgramSocket, &dummy, 1, 0);

	if (result<0)
	{
		  printError();
	}
	i++;
  }
}

/*****************************************************************************\
*
* Raw socket programming. Special privileges are needed under Linux
* It seems to be forbidden (?) under WinXP
*
\*****************************************************************************/
/*****************************************************************************\
*
*
\*****************************************************************************/

int openRawSocket()
{
	rawSocket=socket (AF_INET, SOCK_RAW, IPPROTO_RAW);	// open raw socket 
	printf("Socket opened: %d \n", rawSocket);
	if (rawSocket<0)
	{
		printError();
	}

   // finally, it is very advisable to do a IP_HDRINCL call, to make sure
   // that the kernel knows the header is included in the data, and doesn't
   // insert its own header into the packet before our data 
   // This enables source IP Spoofing...


#ifndef _WIN32  
	char value = 1;
	result=setsockopt (rawSocket, IPPROTO_IP, IP_HDRINCL, &value, sizeof (value));
	if (result < 0)
	{
		printf ("Warning: Cannot set HDRINCL!\n"); 
		printError();
	}
	else
	{
		printf("Enable IP spoofing \n");
	}
#endif  

  return rawSocket;
}

/*****************************************************************************\
*
*
\*****************************************************************************/

void closeRawSocket()
{
#ifdef _WIN32
	closesocket(rawSocket);
#else
    close(rawSocket);
#endif
}



/*****************************************************************************\
*
*
\*****************************************************************************/
void sendTcpSynPacket(char* address, int port)
{
  int					result;
  
  char					datagram[4096];	// this buffer will contain ip header, tcp header,
										// and payload. we'll point an ip header structure
										// at its beginning, and a tcp header structure after
										// that to write the header values into it 
 
  struct ip* 	 		iph;
  struct tcphdr*		tcph;

  struct sockaddr_in 	sin;			// the sockaddr_in containing the dest. address is used
										// in sendto() to determine the datagrams path 

  iph  = (struct ip  *)  datagram;
  tcph = (struct tcphdr *) (datagram + sizeof (struct ip));


/*
    if(( hp = gethostbyname(argv[1])) == NULL){
        if(ip->ip_dst.s_addr = inet_addr(argv[1]) == -1)
             printf("\nERROR: UNKNOWN HOST");
    }
    else
        bcopy(hp->h_addr_list[0], &ip->ip_dst.s_addr,                            hp->h_length);
    printf("\nSending to %s\n", inet_ntoa(ip->ip_dst));
*/


  sin.sin_family      = AF_INET;
  sin.sin_port        = htons (port);						// you byte-order >1byte header values to network
			  												// byte order (not needed on big endian machines)
  sin.sin_addr.s_addr = inet_addr (address);

  memset (datagram, 0, 4096);								// zero out the buffer 

/* we'll now fill in the ip/tcp header values, see above for explanations */
  iph->ip_hl         = 5; /* sizeof *ip >> 2 */				// headerlength
  iph->ip_v          = 4;									// IP Version
  iph->ip_tos        = 0;									// Experimental...
  iph->ip_len        = sizeof (struct ip) + sizeof (*tcph);	// No payload, just the headers
  iph->ip_id         = (u_short)htonl (54321);						// doesn't matter
  iph->ip_off        = 0;						
  iph->ip_ttl        = 255;									// the packet can pass 255 hops
  iph->ip_p          = 6;									// 
  iph->ip_sum        = 0;									// set it to 0 before computing the actual checksum later 
  iph->ip_src.s_addr = inet_addr ("172.18.0.1");			// SYN's can be blindly spoofed 
  iph->ip_dst.s_addr = inet_addr (address);					// destination address
  

  tcph->source       = htons (4900);						// source port: arbitrary
  tcph->dest         = htons (port);						// port to knock
  tcph->seq          = rand();								// in a SYN packet, the sequence is a random */
  tcph->ack_seq      = 0;									// number, and the ack sequence is 0 in the 1st packet 
  tcph->res1         = 0;
  tcph->doff         = 0;									// first and only tcp segment 
  tcph->fin          = 0;									// flag fin: 
  tcph->syn          = 1;									// flag syn: indicate new 
  tcph->rst          = 0;									// flag rst:
  tcph->psh          = 0;									// flag psh: 
  tcph->ack          = 0;									// flag ack:
  tcph->urg          = 0;									// flag urg:
  tcph->window       = (u_short)htonl (65535);				// maximum allowed window size 
  tcph->check        = 0;									// if you set a checksum to zero, your kernel's IP stack
		      												// should fill in the correct checksum during transmission 
  tcph->urg_ptr      = 0;

  iph->ip_sum        = csum ((unsigned short *) datagram, iph->ip_len >> 1);


/*
	result=sendto (	rawSocket,						// our socket 
					datagram,						// the buffer containing headers and data 
					iph->ip_len,					// total length of our datagram 
					0,								// routing flags, normally always 0 
		  			(struct sockaddr *) &sin,		// socket addr, just like in 
		  			sizeof (sin));					// socket addr length
*/

	result=sendto (	rawSocket,						// our socket 
					datagram+20,						// the buffer containing headers and data 
					20,					// total length of our datagram 
					0,								// routing flags, normally always 0 
		  			(struct sockaddr *) &sin,		// socket addr, just like in 
		  			sizeof (sin));					// socket addr length


  if (result < 0)		
  {
    printf ("error %d: \n", result, errno);
    printError();
  }
  else
  {
    printf (".");
  }

}


/*****************************************************************************\
*
*
\*****************************************************************************/
/*
void ping(char* address)
{
     int icmp_socket;
     int ret;
     struct icmp *p, *rp;
     struct sockaddr_in addr, from;
     struct ip *ip;
     char buffer[1024];
     socklen_t sl;
     int hlen;
     int ping_pkt_size;
 		
     icmp_socket = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
     if(icmp_socket < 0)
     {
         printError();
         exit(0);
     }
 
     memset(buffer, 0x00, 1024);
 
     p = (struct icmp *)buffer;
     p->icmp_type=ICMP_ECHO;
     p->icmp_code=0;
     p->icmp_cksum=0;
     p->icmp_seq=15;
     p->icmp_id=getpid();
 
     p->icmp_cksum = csum((u_short *)p, 500);
     memset(&addr, 0, sizeof(0));
     addr.sin_addr.s_addr = inet_addr(address);
     addr.sin_family = AF_INET;
 
 
     ret=sendto(icmp_socket,p,sizeof(*p),MSG_DONTWAIT,(struct sockaddr *)&addr, sizeof(addr));
     if (ret< 0)
     {
         printError();
     }
 	
     sl=sizeof(from);
     ret=recvfrom(icmp_socket, buffer, 1024, 0, (struct sockaddr *)&from, &sl);
     if (ret<0)
     {
       printf("%d %d %d\n", ret, errno, EAGAIN);
       perror("recvfrom error : ");
     }

     ip = (struct ip *)buffer;
     hlen = ip->ip_hl*4;
     rp = (struct icmp *)(buffer+hlen);
     printf("reply from %s\n", inet_ntoa(from.sin_addr));
     printf("Type : %d \n", rp->icmp_type);
     printf("Code : %d \n", rp->icmp_code);
     printf("Seq  : %d \n", rp->icmp_seq);
     printf("Iden : %d \n", rp->icmp_id);
}
*/


/*********************************************************************************\
*
\*********************************************************************************/

bool loadSettings(char* fileName)
{
	FILE	*file;
	bool	error;
	char	dummy[STRING_SIZE];
	int		i;

	error=false;

	if (fileName==NULL)
	{
		printf("Opening file %s\n", DEFAULTFILENAME);
		file=fopen(DEFAULTFILENAME, "rt");
	}
	else
	{
		printf("Opening file %s\n", fileName);
		file=fopen(fileName, "rt");
	}

	if (file != NULL)
	{
		fscanf(file, "%s\n", destinationPc);

		fscanf(file, "%d\n", &destinationPort);
		fscanf(file, "%s\n", dummy);

		i=0;
		while (i<MAX_SEQUENCE)
		{
			knocks[i]=dummy[i]-48;
			if ((knocks[i]<MIN_SEQUENCE_KNOCKS) || (knocks[i]>MAX_SEQUENCE_KNOCKS))
			{
				error=true;
			}
			i++;
		}

		fscanf(file, "%d\n", &knockTimeWindow);
		fclose(file);
	}
	else
	{
		if (fileName==NULL)
		{
			printf("Cannot open file %s\n", DEFAULTFILENAME);
		}
		else
		{
			printf("Cannot open file %s\n", fileName);
		}

		error=true;
	}

	return error;
}

/*********************************************************************************\
*
\*********************************************************************************/
bool getSettings(int argc, char* argv[])
{
	bool error;
	int  i;

	error=false;

	strncpy(destinationPc, argv[1], STRING_SIZE);
    destinationPort=atoi(argv[2]);
	knockTimeWindow=atoi(argv[4]);

	i=0;
	while (i<MAX_SEQUENCE)
	{
		knocks[i]=argv[3][i]-48;
		if ((knocks[i]<MIN_SEQUENCE_KNOCKS) || (knocks[i]>MAX_SEQUENCE_KNOCKS))
		{
			error=true;
		}
		i++;
	}

	return error;
}


/*********************************************************************************\
*
\*********************************************************************************/

bool checkServerAddress(char* serverAddress)
{
	bool			error;
	unsigned int	i;
	bool			isIp;
	hostent*		host;

	error=false;

	isIp=true;
	i=0;
	while (i<strlen(serverAddress))
	{
		if (((serverAddress[i]<'0') || (serverAddress[i]>'9')) &&
			(serverAddress[i]!='.'))
		{
			isIp=false;
		}
		i++;
	}

	if (!isIp)
	{
		printf("Translating Servername %s ", serverAddress);
		host=gethostbyname(serverAddress);
		if (host!=NULL)
		{
			sprintf(serverAddress, "%d.%d.%d.%d", host->h_addr_list[0][0],
												  host->h_addr_list[0][1],
												  host->h_addr_list[0][2],
												  host->h_addr_list[0][3]);
			printf("to IP number: result %s\n", serverAddress);
		}
		else
		{
			printf("error obtaining IP address\n", serverAddress);
			error=true;
		}
	}


	return error;
}

/*****************************************************************************\
*
*
\*****************************************************************************/
main(int argc, char* argv[])
{
  int  i;
  bool error;


  printf("#######################################################################\n");
  printf("#                       O P E N   S E S A M Y                         #\n");
  printf("#######################################################################\n");

  error=false;

  if (argc==5)
  {
	  error=getSettings(argc, argv);
  }
  else if (argc==1)
  {
	  error=loadSettings(NULL);
  }
  else if (argc==2)
  {
	  error=loadSettings(argv[1]);
  }
  else
  {
	  error=true;
  }



  if (!error)
  {
	initialise();

	checkServerAddress(destinationPc);

    printf("IP Address: %s, port %d sequence: ", destinationPc, destinationPort);
	i=0;
	while (i<MAX_SEQUENCE)
	{
		printf("%d ", knocks[i]);
		i++;
	}
	printf("\n");

/*
    openRawSocket();
    
    sendTcpSynPacket(argv[1], port);    
  
    closeRawSocket();
*/

	if (!error)
	{

		openDgramSocket(destinationPc, destinationPort);



		i=0;
		while (i<MAX_SEQUENCE)
		{
			printf("Knocking...\n");
			sendDgram(knocks[i]);

			if (i!=MAX_SEQUENCE-1)
			{
#ifdef _WIN32
				Sleep(knockTimeWindow);
#else
				usleep(knockTimeWindow*1000);
#endif
			}

			i++;
		}
	}

	printf("Secured port should be opened now for %d seconds..\n", knockTimeWindow/1000);
    printf("#######################################################################\n");




	wrapUp();

 
  }

  if (error)
  {
    printf("Mmmm...\n");
	printf("Usage 1: PortKnocker server knockport sequence timewindow\n");
	printf("           server      : IP of server to knock\n");
	printf("           knockport   : port to knock\n");
	printf("           sequence    : sequence of knocks, like 123\n");
	printf("           timewindow  : timeslot in ms for the knock sequences\n");
	printf("           E.g. PortKnocker 192.168.1.1 22 1234 3000");
	printf("\n\nUsage 2: PortKnocker\n");
	printf("           A file called knocksettings.txt should contain \n");
	printf("           the values on separate lines, e.g.:\n");
	printf("             192.168.1.1\n");
	printf("             22\n");
	printf("             1234\n");
	printf("             3000\n");
	printf("\n\nUsage 3: PortKnocker filename\n");
	printf("             Same as 2, except filename can be specified\n");
    printf("#######################################################################\n");
  }



    
  return 0;
}




// Stuff not to forget...
// This stuff gives problems under WinXP
/* 
    openRawSocket();
    
    sendTcpSynPacket(argv[1], port);    
    sendTcpSynPacket(argv[1], port);    
    sendTcpSynPacket(argv[1], port);  
	

#ifdef _WIN32
	Sleep(6000);
#else
    usleep(6000000);
#endif
	
	sendTcpSynPacket(argv[1], port);    
    sendTcpSynPacket(argv[1], port);    
  
    closeRawSocket();
*/


//    ping(argv[1]); 

